# -*- coding: utf-8 -*-

import argparse
import sys

from osgeo import gdal

from threedidepth.calculate import calculate_waterdepth
from threedidepth.calculate import MODE_LIZARD
from threedidepth.calculate import MODE_CONSTANT


def threedidepth(*args):
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "gridadmin_path", metavar="gridadmin", help="path to gridadmin.h5 file"
    )
    parser.add_argument(
        "results_3di_path",
        metavar="results_3di",
        help="path to (aggregate_)results_3di.nc file",
    )
    parser.add_argument(
        "dem_path", metavar="dem", help="path to bathymetry file"
    )
    parser.add_argument(
        "waterdepth_path",
        metavar="waterdepth",
        help="path to resulting geotiff"
    )
    parser.add_argument(
        "-s",
        "--steps",
        nargs="+",
        type=int,
        dest="calculation_steps",
        help="simulation result step(s)",
    )
    parser.add_argument(
        "-c",
        "--constant",
        action="store_true",
        help="disable interpolation and use constant waterlevel per grid cell",
    )
    parser.add_argument(
        "-p",
        "--progress",
        action="store_const",
        dest="progress_func",
        const=gdal.TermProgress_nocb,
        help="Show progress.",
    )
    parser.add_argument(
        "-n",
        "--netcdf",
        action="store_true",
        help="export the waterdepth as a netcdf"
    )
    kwargs = vars(parser.parse_args())
    if kwargs.pop("constant"):
        kwargs["mode"] = MODE_CONSTANT
    else:
        kwargs["mode"] = MODE_LIZARD
    calculate_waterdepth(**kwargs)


if __name__ == '__main__':
    threedidepth(sys.argv)



# -*- coding: utf-8 -*-

import pytest

import numpy as np

from threedidepth import morton


def test_morton_array():
    result = morton.morton_array((8, 8)).tolist()
    expected = [[ 0,  1,  4,  5, 16, 17, 20, 21],  # noqa: E201
                [ 2,  3,  6,  7, 18, 19, 22, 23],  # noqa: E201
                [ 8,  9, 12, 13, 24, 25, 28, 29],  # noqa: E201
                [10, 11, 14, 15, 26, 27, 30, 31],
                [32, 33, 36, 37, 48, 49, 52, 53],
                [34, 35, 38, 39, 50, 51, 54, 55],
                [40, 41, 44, 45, 56, 57, 60, 61],
                [42, 43, 46, 47, 58, 59, 62, 63]]
    assert result == expected
    assert morton.morton_array((2,)).tolist() == [0, 1]
    assert morton.morton_array((1, 2)).tolist() == [[0, 1]]
    with pytest.raises(ValueError):
        morton.morton_array((2 ** 32, 2 ** 32))


def test_group():
    array = np.array([0.6, 0.6, 0.6, 0.5, 0.7, 0.5, 0.7, 0.5, 0.7])
    expected = [[3, 5, 7], [0, 1, 2], [4, 6, 8]]
    result = [i.tolist() for i in morton.group(array)]
    assert result == expected


def test_rasterize():
    # single point
    points = np.array([[0, 0]])
    expected = [[0]]
    result = morton.rasterize(points)[0].tolist()
    assert result == expected

    # horizontal
    points = np.array([[0, 0], [1, 0]])
    expected = [[0, 1]]
    result = morton.rasterize(points)[0].tolist()
    assert result == expected

    # vertical
    points = np.array([[0, 0], [1, 0]])
    expected = [[0, 1]]
    result = morton.rasterize(points)[0].tolist()
    assert result == expected

    # quadtree
    points = np.array([[1, 1], [3, 1], [1, 3], [3, 3], [6, 2]])
    expected = [
        [2, 3, 4],
        [0, 1, 5],
    ]
    result = morton.rasterize(points)[0].tolist()
    assert result == expected


def test_reorder():
    points = np.array([[1, 3], [3, 1], [6, 2], [3, 3], [1, 1]])
    s1 = np.array([1, 2, 3, 4, 5])

    result_points, result_s1 = morton.reorder(points, s1)

    expected_points = [[1, 3], [3, 3], [1, 1], [3, 1], [6, 2]]
    expected_s1 = [1, 4, 5, 2, 3]

    assert result_points.tolist() == expected_points
    assert result_s1.tolist() == expected_s1



# -*- coding: utf-8 -*-

from contextlib import contextmanager
from unittest import mock
import random
import string

from numpy.testing import assert_allclose
from osgeo import gdal
from osgeo import osr
from pytest import fixture
from pytest import mark
from pytest import raises
import numpy as np

from threedidepth.calculate import BaseCalculator
from threedidepth.calculate import GeoTIFFConverter
from threedidepth.calculate import MODE_CONSTANT
from threedidepth.calculate import MODE_CONSTANT_S1
from threedidepth.calculate import MODE_COPY
from threedidepth.calculate import MODE_LINEAR
from threedidepth.calculate import MODE_LINEAR_S1
from threedidepth.calculate import MODE_LIZARD
from threedidepth.calculate import MODE_LIZARD_S1
from threedidepth.calculate import MODE_NODGRID
from threedidepth.calculate import NetcdfConverter
from threedidepth.calculate import ProgressClass
from threedidepth.calculate import SUBSET_2D_OPEN_WATER
from threedidepth.calculate import calculate_waterdepth
from threedidepth.calculate import calculator_classes

RD = osr.GetUserInputAsWKT("EPSG:28992")
NDV = -9  # no data value of the test dem
NLV = -9999.  # "no level value"


def random_slug(size):
    """Return a random slug of length `size`.

    The returned slug consists of digits and lowercase letters.
    """
    chars = string.ascii_lowercase + string.digits
    return "".join(random.choice(chars) for _ in range(size))


@contextmanager
def vsimem_path():
    """ An autoremoving temporary path in vsimem. """
    while True:
        vsimem_path = "/vsimem/" + random_slug(6)
        if gdal.VSIStatL(vsimem_path) is None:
            break
    yield vsimem_path
    gdal.Unlink(vsimem_path)


@fixture
def target_path():
    with vsimem_path() as path:
        yield path


@fixture(scope="module")
def source_path(request):
    """Provides a GeoTIFF file in the vsimem virtual filesystem.

    Can be parametrized to add the creation option "tiled=yes".
    """
    bands, height, width, data_type = 1, 128, 512, gdal.GDT_Float32
    array = np.arange(height * width).reshape(height, width)

    source = gdal.GetDriverByName("mem").Create(
        "", width, height, bands, data_type,
    )
    source.SetGeoTransform((0, 1, 0, 128, 0, -1))
    source.SetProjection(RD)
    source_band = source.GetRasterBand(1)
    source_band.SetNoDataValue(-9)
    source_band.WriteArray(array)

    options = ["compress=deflate"]
    if getattr(request, "param", False):
        options.append("tiled=yes")

    with vsimem_path() as path:
        gdal.GetDriverByName("gtiff").CreateCopy(path, source, options=options)
        yield path


@fixture
def admin():
    result_admin = mock.Mock()
    import_path = "threedidepth.calculate.ResultAdmin"
    with mock.patch(import_path) as ResultAdmin:
        ResultAdmin.return_value = result_admin
        yield result_admin


def test_progress_class():
    progress_func = mock.Mock()
    calculation_steps = [10, 20]
    progress_class = ProgressClass(
        calculation_steps=[10, 20],
        progress_func=progress_func,
    )
    for i, calculation_step in progress_class:
        assert calculation_step == calculation_steps[i]
        progress_class(0.5)
        assert progress_func.called_with(i + 0.25)
        progress_class(1.0)
        assert progress_func.called_with(i + 0.50)


@mark.parametrize("source_path", [False, True], indirect=True)
def test_geotiff_converter(source_path, target_path):
    progress_func = mock.Mock()
    band_count = 3
    converter_kwargs = {
        "source_path": source_path,
        "target_path": target_path,
        "band_count": band_count,
        "progress_func": progress_func,
    }

    def calculator(indices, values, no_data_value):
        """Return input values unmodified."""
        return values

    with GeoTIFFConverter(**converter_kwargs) as converter:
        converter.convert_using(calculator=calculator, band=0)

        # check progess func calls
        progress_func_values = [r[0][0] for r in progress_func.call_args_list]
        assert len(progress_func_values) == len(converter)
        assert progress_func_values[0] < 1
        assert progress_func_values[-1] == 1
        assert progress_func_values == sorted(progress_func_values)

        # convert another band
        converter.convert_using(calculator=calculator, band=2)

    # check target
    source = gdal.Open(source_path)
    source_array = source.ReadAsArray()
    source_band = source.GetRasterBand(1)
    source_no_data_value = source_band.GetNoDataValue()
    source_block_size = source_band.GetBlockSize()

    target = gdal.Open(target_path)
    target_array = target.ReadAsArray()
    target_bands = [target.GetRasterBand(i + 1) for i in range(band_count)]
    target_no_data_values = [b.GetNoDataValue() for b in target_bands]
    target_block_sizes = [b.GetBlockSize() for b in target_bands]

    assert source.GetGeoTransform() == target.GetGeoTransform()
    assert source.GetProjection() == target.GetProjection()
    assert all(source_block_size == v for v in target_block_sizes)
    assert all(source_no_data_value == v for v in target_no_data_values)
    assert len(target_array) == band_count
    assert np.equal(source_array, target_array[[0, 2]]).all()
    assert np.equal(0, target_array[[1]]).all()


def test_tiff_converter_existing_target(tmpdir):
    target_path = tmpdir.join("target.tif")
    target_path.ensure(file=True)  # "touch" the file
    with raises(OSError, match="exists"):
        GeoTIFFConverter(
            source_path=None, target_path=target_path, progress_func=None)


@mark.parametrize("source_path", [False, True], indirect=True)
def test_netcdf_converter(source_path, tmp_path, admin):
    admin.result_type = "raw"
    admin.get_timestamps.return_value = np.array([1, -1])
    admin.get_time_units.return_value = b'seconds since 2021-03-02 10:00:00'
    admin.model_slug = "my_model_slug"
    admin.epsg_code = "28992"
    admin.variable = "s1"
    target_path = str(tmp_path / "waterdepth.nc")

    progress_func = mock.Mock()
    converter_kwargs = {
        "source_path": source_path,
        "target_path": target_path,
        "result_admin": admin,
        "calculation_steps": [1, -1],
        "progress_func": progress_func,
    }

    def calculator(indices, values, no_data_value):
        """Return input values unmodified."""
        return values

    with NetcdfConverter(**converter_kwargs) as converter:
        converter.convert_using(calculator, band=0)
        converter.convert_using(calculator, band=1)

        assert 2 * len(converter) == len(progress_func.call_args_list)
        assert progress_func.call_args_list[0][0][0] < 1
        assert progress_func.call_args_list[-1][0][0] == 1

    source = gdal.Open(source_path)
    target = gdal.Open(target_path)

    assert np.equal(source.ReadAsArray(), target.ReadAsArray()).all()
    assert source.GetGeoTransform() == target.GetGeoTransform()


def test_calculate_waterdepth_wrong_mode():
    with raises(ValueError, match="Unknown mode:"):
        calculate_waterdepth(
            gridadmin_path="dummy",
            results_3di_path='dummy',
            dem_path="dummy",
            waterdepth_path="dummy",
            mode="wrong",
        )


def test_calculate_waterdepth_wrong_step(admin):
    admin.calculation_steps = 2
    with raises(AssertionError, match="Maximum calculation step is"):
        calculate_waterdepth(
            gridadmin_path="dummy",
            results_3di_path='dummy',
            dem_path="dummy",
            waterdepth_path="dummy",
            calculation_steps=[3],
            mode=MODE_COPY,
        )


def test_calculate_waterdepth(source_path, target_path, admin):
    admin.calculation_steps = 2
    with mock.patch("threedidepth.calculate.fix_gridadmin"):
        calculate_waterdepth(
            gridadmin_path="dummy",
            results_3di_path='dummy',
            dem_path=source_path,
            waterdepth_path=target_path,
            mode=MODE_COPY,
        )


def test_depth_from_water_level():
    dem = np.array([[7.0, 2.0], [3.0, 4.0]])
    fillvalue = 7.0
    waterlevel = np.array([[4.0, NDV], [4.0, 4.0]])
    depth = BaseCalculator._depth_from_water_level(
        dem=dem, fillvalue=fillvalue, waterlevel=waterlevel,
    )
    assert depth[0, 0] == fillvalue  # because dem is fillvalue
    assert depth[0, 1] == fillvalue  # because waterlevel is no data
    assert depth[1, 0] == 1.0        # really, a depth
    assert depth[1, 1] == fillvalue  # because depth <= 0


# some interpolated values as expected results for the linear mode
EL1 = 4.0 / 4 + NLV * 3 / 4
EL2 = 6.0 / 3 + NLV * 2 / 3
EL3 = 5.0 / 2 + NLV / 4 + 4.0 / 4
EL4 = 6.0 / 3 + NLV / 6 + 5.0 / 2
data = (
    (
        MODE_COPY,
        np.array([[6.0, 6.0, 6.0, 6.0],
                  [6.0, 6.0, 6.0, 6.0],
                  [6.0, 6.0, 6.0, NDV]]),
    ),
    (
        MODE_NODGRID,
        np.array([[2, 2, 3, 3],
                  [0, 5, 6, 7],
                  [5, 5, 8, 9]]),
    ),
    (
        MODE_CONSTANT_S1,
        np.array([[NLV, NLV, NLV, NLV],
                  [NLV, 5.0, 6.0, 7.0],
                  [5.0, 5.0, 8.0, 9.0]]),
    ),
    (
        MODE_LINEAR_S1,
        np.array([[EL1, EL2, EL2, NLV],
                  [EL3, EL4, 6.0, 7.0],
                  [NLV, NLV, 8.0, 9.0]]),
    ),
    (
        MODE_LIZARD_S1,
        np.array([[NLV, NLV, NLV, NLV],
                  [NLV, 5.4, 6.0, 7.0],
                  [5.0, 5.0, 8.0, 9.0]]),
    ),
)


@mark.parametrize("mode, expected", data)
def test_calculators(mode, expected, admin):
    nodgrid = np.array([
        [1, 1, 2, 2, 3, 3],
        [1, 1, 2, 2, 3, 3],
        [4, 4, 0, 5, 6, 7],
        [4, 4, 5, 5, 8, 9],
    ])  # note that 0 means no grid cell at all in threedigrid
    geo_transform = (20, 2, 0, 48, 0, -2)  # bbox: (20, 40, 32, 48)
    s1 = np.array([[1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 8.0, 9.0]])
    s1[:, :3] = -9999.0
    ids = np.arange(1, 10)
    coordinates = np.array([
        [22, 26, 30, 22, 26, 29, 31, 29, 31],
        [46, 46, 46, 42, 42, 43, 43, 41, 41],
    ])

    # request indices and selection of nodgrid
    indices = ((1, 2), (4, 6))  # request bbox (24, 42), (32, 48)

    # dem values
    values = np.array([
        [6.0, 6.0, 6.0, 6.0],
        [6.0, 6.0, 6.0, 6.0],
        [6.0, 6.0, 6.0, NDV],
    ])

    # prepare gridadmin uses
    admin.cells.get_nodgrid.return_value = nodgrid[tuple(map(slice, *indices))]
    admin.variable = "s1"
    if mode in (MODE_CONSTANT_S1):
        data = {"id": ids, "s1": s1}
        admin.nodes.subset().timeseries().only().data = data
    if mode in (MODE_LINEAR_S1):
        data = {"coordinates": coordinates, "s1": s1}
        admin.nodes.subset().timeseries().only().data = data
    if mode in (MODE_LIZARD_S1):
        admin.nodes.subset().timeseries().only.side_effect = [
            mock.Mock(data={"id": ids, "s1": s1}),
            mock.Mock(data={"coordinates": coordinates, "s1": s1}),
        ]

    indexes = slice(7, 8)
    calculator_kwargs = {
        "result_admin": admin,
        "calculation_step": indexes.start,
        "dem_shape": nodgrid.shape,
        "dem_geo_transform": geo_transform,
    }

    call_kwargs = {
        "no_data_value": NDV,
        "values": values,
        "indices": indices,
    }

    CalculatorClass = calculator_classes[mode]
    with CalculatorClass(**calculator_kwargs) as calculator:
        result = calculator(**call_kwargs)
    assert_allclose(result, expected)

    # check gridadmin uses
    if mode in (MODE_NODGRID, MODE_CONSTANT_S1, MODE_LIZARD_S1):
        # get_nodgrid() wants (1, 2), (4, 6) flipped & transposed
        admin.cells.get_nodgrid.assert_called_once_with(
            [2, 0, 6, 3],
            subset_name=SUBSET_2D_OPEN_WATER,
        )
    if mode in (MODE_LINEAR_S1):
        admin.nodes.subset.assert_called_with(SUBSET_2D_OPEN_WATER)
        admin.nodes.subset().timeseries.assert_called_with(indexes=indexes)
        admin.nodes.subset().timeseries().only.assert_called_with(
            "s1", "coordinates",
        )
    if mode in (MODE_CONSTANT_S1):
        admin.nodes.subset.assert_called_with(SUBSET_2D_OPEN_WATER)
        admin.nodes.subset().timeseries.assert_called_with(indexes=indexes)
        admin.nodes.subset().timeseries().only.assert_called_with(
            "s1", "id",
        )
    if mode in (MODE_LIZARD_S1):
        admin.nodes.subset.assert_called_with(SUBSET_2D_OPEN_WATER)
        admin.nodes.subset().timeseries.assert_called_with(indexes=indexes)
        admin.nodes.subset().timeseries().only.assert_has_calls(
            [mock.call("s1", "id"), mock.call("s1", "coordinates")]
        )


@fixture
def depthmock():
    import_path = (
        "threedidepth.calculate.BaseCalculator._depth_from_water_level"
    )
    with mock.patch(import_path) as depthmock:
        yield depthmock


mode = (
    MODE_CONSTANT,
    MODE_LINEAR,
    MODE_LIZARD,
)


@mark.parametrize("mode", mode)
def test_depth_calculators(depthmock, mode):
    CalculatorClass = calculator_classes[mode]
    BaseClass = CalculatorClass.__bases__[0]
    depthmock.return_value = 5
    import_path = "threedidepth.calculate." + BaseClass.__name__ + ".__call__"
    with mock.patch(import_path) as callmock:
        callmock.return_value = 4
        calculator = CalculatorClass("", "", "", "")
        depth = calculator(1, 2, 3)

    callmock.assert_called_once_with(1, 2, 3)
    depthmock.assert_called_once_with(dem=2, fillvalue=3, waterlevel=4)
    assert depth == 5


def test_calculator_not_implemented():
    with raises(NotImplementedError):
        BaseCalculator("", "", "", "")("", "", "")



# -*- coding: utf-8 -*-

import h5py
import numpy as np
from threedigrid.admin.nodes.subsets import NODE_TYPE__IN_SUBSETS


def fix_gridadmin(gridadmin_path):
    """Fix older type gridadmin files.
    """
    with h5py.File(gridadmin_path, "r") as h5py_file:
        node_group = h5py_file["nodes"]
        if "pixel_coords" in node_group:
            return

    with h5py.File(gridadmin_path, "r+") as h5py_file:
        nodk = h5py_file["grid_coordinate_attributes"]["nodk"].value
        nodm = h5py_file["grid_coordinate_attributes"]["nodm"].value
        nodn = h5py_file["grid_coordinate_attributes"]["nodn"].value
        ip = h5py_file["grid_coordinate_attributes"]["ip"].value
        jp = h5py_file["grid_coordinate_attributes"]["jp"].value
        node_types = h5py_file["nodes"]["node_type"].value
        pixel_coords = np.full((4, nodk.shape[0]), -9999, dtype="int")
        for node_type_subset in NODE_TYPE__IN_SUBSETS["2D_ALL"]:
            mask = node_types == node_type_subset
            pixel_coords[0, mask] = ip[0, nodm[mask] - 1, nodk[mask] - 1] - 1
            pixel_coords[1, mask] = jp[0, nodn[mask] - 1, nodk[mask] - 1] - 1
            pixel_coords[2, mask] = ip[3, nodm[mask] - 1, nodk[mask] - 1]
            pixel_coords[3, mask] = jp[3, nodn[mask] - 1, nodk[mask] - 1]
        node_group = h5py_file["nodes"]
        node_group.create_dataset(
            "pixel_coords", data=pixel_coords, dtype="int"
        )



# -*- coding: utf-8 -*-

from itertools import product
from os import path

from osgeo import gdal
from osgeo import osr
from scipy.interpolate import LinearNDInterpolator
from scipy.spatial import qhull
import h5netcdf.legacyapi as netCDF4
import h5py
import numpy as np

from threedigrid.admin.gridresultadmin import GridH5ResultAdmin
from threedigrid.admin.gridresultadmin import GridH5AggregateResultAdmin
from threedigrid.admin.constants import SUBSET_2D_OPEN_WATER
from threedigrid.admin.constants import NO_DATA_VALUE
from threedidepth.fixes import fix_gridadmin
from threedidepth import morton

MODE_COPY = "copy"
MODE_NODGRID = "nodgrid"
MODE_CONSTANT_S1 = "constant-s1"
MODE_LINEAR_S1 = "linear-s1"
MODE_LIZARD_S1 = "lizard-s1"
MODE_CONSTANT = "constant"
MODE_LINEAR = "linear"
MODE_LIZARD = "lizard"


class BaseCalculator:
    """Depth calculator using constant waterlevel in a grid cell.

    Args:
        result_admin (ResultAdmin): ResultAdmin instance.
        calculation_step (int): Calculation step.
        dem_shape (int, int): Shape of the dem array.
        dem_geo_transform: (tuple) Geo_transform of the dem.
    """

    PIXEL_MAP = "pixel_map"
    LOOKUP_S1 = "lookup_s1"
    INTERPOLATOR = "interpolator"
    DELAUNAY = "delaunay"

    def __init__(
        self, result_admin, calculation_step, dem_shape, dem_geo_transform,
    ):
        self.ra = result_admin
        self.calculation_step = calculation_step
        self.dem_shape = dem_shape
        self.dem_geo_transform = dem_geo_transform

    def __call__(self, indices, values, no_data_value):
        """Return result values array.

        Args:
            indices (tuple): ((i1, j1), (i2, j2)) subarray indices
            values (array): source values for the calculation
            no_data_value (scalar): source and result no_data_value

        Override this method to implement a calculation. The default
        implementation is to just return the values, effectively copying the
        source.

        Note that the no_data_value for the result has to correspond to the
        no_data_value argument.
        """
        raise NotImplementedError

    @staticmethod
    def _depth_from_water_level(dem, fillvalue, waterlevel):
        # determine depth
        depth = np.full_like(dem, fillvalue)
        dem_active = dem != fillvalue
        waterlevel_active = waterlevel != NO_DATA_VALUE
        active = dem_active & waterlevel_active
        depth_1d = waterlevel[active] - dem[active]

        # paste positive depths only
        negative_1d = depth_1d <= 0
        depth_1d[negative_1d] = fillvalue
        depth[active] = depth_1d

        return depth

    @property
    def indexes(self):
        return slice(self.calculation_step, self.calculation_step + 1)

    @property
    def lookup_s1(self):
        """
        Return the lookup table to find waterlevel by cell id.

        Both cells outside any defined grid cell and cells in a grid cell that
        are currently not active ('no data') will return the NO_DATA_VALUE as
        defined in threedigrid.
        """
        try:
            return self.cache[self.LOOKUP_S1]
        except KeyError:
            nodes = self.ra.nodes.subset(SUBSET_2D_OPEN_WATER)
            timeseries = nodes.timeseries(indexes=self.indexes)
            data = timeseries.only(self.ra.variable, "id").data
            lookup_s1 = np.full((data["id"]).max() + 1, NO_DATA_VALUE)
            lookup_s1[data["id"]] = data[self.ra.variable][0]
            self.cache[self.LOOKUP_S1] = lookup_s1
        return lookup_s1

    @property
    def interpolator(self):
        try:
            return self.cache[self.INTERPOLATOR]
        except KeyError:
            nodes = self.ra.nodes.subset(SUBSET_2D_OPEN_WATER)
            timeseries = nodes.timeseries(indexes=self.indexes)
            data = timeseries.only(self.ra.variable, "coordinates").data
            points = data["coordinates"].transpose()
            values = data[self.ra.variable][0]
            interpolator = LinearNDInterpolator(
                points, values, fill_value=NO_DATA_VALUE
            )
            self.cache[self.INTERPOLATOR] = interpolator
            return interpolator

    @property
    def delaunay(self):
        """
        Return a (delaunay, s1) tuple.

        `delaunay` is a qhull.Delaunay object, and `s1` is an array of
        waterlevels for the corresponding delaunay vertices.
        """
        try:
            return self.cache[self.DELAUNAY]
        except KeyError:
            nodes = self.ra.nodes.subset(SUBSET_2D_OPEN_WATER)
            timeseries = nodes.timeseries(indexes=self.indexes)
            data = timeseries.only(self.ra.variable, "coordinates").data
            points = data["coordinates"].transpose()
            s1 = data[self.ra.variable][0]

            # reorder a la lizard
            points, s1 = morton.reorder(points, s1)

            delaunay = qhull.Delaunay(points)
            self.cache[self.DELAUNAY] = delaunay, s1
            return delaunay, s1

    def _get_nodgrid(self, indices):
        """Return node grid.

        Args:
            indices (tuple): ((i1, j1), (i2, j2)) subarray indices
        """
        (i1, j1), (i2, j2) = indices

        # note that get_nodgrid() starts counting rows from the bottom
        h = self.dem_shape[0]
        i1, i2 = h - i2, h - i1

        # note that get_nodgrid() expects a columns-first bbox
        return self.ra.cells.get_nodgrid(
            [j1, i1, j2, i2], subset_name=SUBSET_2D_OPEN_WATER
        )

    def _get_points(self, indices):
        """Return points array.

        Args:
            indices (tuple): ((i1, j1), (i2, j2)) subarray indices
        """
        (i1, j1), (i2, j2) = indices
        local_ji = np.mgrid[i1:i2, j1:j2].reshape(2, -1)[::-1].transpose()
        p, a, b, q, c, d = self.dem_geo_transform
        return local_ji * [a, d] + [p + 0.5 * a, q + 0.5 * d]

    def __enter__(self):
        self.cache = {}
        return self

    def __exit__(self, *args):
        self.cache = None


class CopyCalculator(BaseCalculator):
    def __call__(self, indices, values, no_data_value):
        """Return input values unmodified."""
        return values


class NodGridCalculator(BaseCalculator):
    def __call__(self, indices, values, no_data_value):
        """Return node grid."""
        return self._get_nodgrid(indices)


class ConstantLevelCalculator(BaseCalculator):
    def __call__(self, indices, values, no_data_value):
        """Return waterlevel array."""
        return self.lookup_s1[self._get_nodgrid(indices)]


class LinearLevelCalculator(BaseCalculator):
    def __call__(self, indices, values, no_data_value):
        """Return waterlevel array."""
        points = self._get_points(indices)
        return self.interpolator(points).reshape(values.shape)


class LizardLevelCalculator(BaseCalculator):
    def __call__(self, indices, values, no_data_value):
        """ Return waterlevel array.

        This uses both the grid layout from the constant level method and the
        triangulation from the linear method.

        Interpolation is used to determine the waterlevel for a result cell if
        all of the following requirements are met:
        - The point is inside a grid cell
        - The point is inside the triangulation
        - The sum of weights of active (not 'no data' nodes) is more than half
          of the total weight of all nodes. Only active nodes are included in
          the interpolation.

        In all other cases, the waterlevel from the constant level method is
        used."""
        # start with the constant level result
        nodgrid = self._get_nodgrid(indices).ravel()
        level = self.lookup_s1[nodgrid]

        # determine result raster cell centers and in which triangle they are
        points = self._get_points(indices)
        delaunay, s1 = self.delaunay
        simplices = delaunay.find_simplex(points)

        # determine which points will use interpolation
        in_gridcell = nodgrid != 0
        in_triangle = simplices != -1
        in_interpol = in_gridcell & in_triangle
        points = points[in_interpol]

        # get the nodes and the transform for the corresponding triangles
        transform = delaunay.transform[simplices[in_interpol]]
        vertices = delaunay.vertices[simplices[in_interpol]]

        # calculate weight, see print(spatial.Delaunay.transform.__doc__) and
        # Wikipedia about barycentric coordinates
        weight = np.empty(vertices.shape)
        weight[:, :2] = np.sum(
            transform[:, :2] * (points - transform[:, 2])[:, np.newaxis], 2
        )
        weight[:, 2] = 1 - weight[:, 0] - weight[:, 1]

        # set weight to zero when for inactive nodes
        nodelevel = s1[vertices]
        weight[nodelevel == NO_DATA_VALUE] = 0

        # determine the sum of weights per result cell
        weight_sum = weight.sum(axis=1)

        # further subselect points suitable for interpolation
        suitable = weight_sum > 0.5
        weight = weight[suitable] / weight_sum[suitable][:, np.newaxis]
        nodelevel = nodelevel[suitable]

        # combine weight and nodelevel into result
        in_interpol_and_suitable = in_interpol.copy()
        in_interpol_and_suitable[in_interpol] &= suitable
        level[in_interpol_and_suitable] = np.sum(weight * nodelevel, axis=1)
        return level.reshape(values.shape)


class ConstantLevelDepthCalculator(ConstantLevelCalculator):
    def __call__(self, indices, values, no_data_value):
        """Return waterdepth array."""
        waterlevel = super().__call__(indices, values, no_data_value)
        return self._depth_from_water_level(
            dem=values, fillvalue=no_data_value, waterlevel=waterlevel
        )


class LinearLevelDepthCalculator(LinearLevelCalculator):
    def __call__(self, indices, values, no_data_value):
        """Return waterdepth array."""
        waterlevel = super().__call__(indices, values, no_data_value)
        return self._depth_from_water_level(
            dem=values, fillvalue=no_data_value, waterlevel=waterlevel
        )


class LizardLevelDepthCalculator(LizardLevelCalculator):
    def __call__(self, indices, values, no_data_value):
        """Return waterdepth array."""
        waterlevel = super().__call__(indices, values, no_data_value)
        return self._depth_from_water_level(
            dem=values, fillvalue=no_data_value, waterlevel=waterlevel
        )


class GeoTIFFConverter:
    """Convert tiff, applying a calculating function to the data.

    Args:
        source_path (str): Path to source GeoTIFF file.
        target_path (str): Path to target GeoTIFF file.
        band_count (int): Number of bands in the target file.
        progress_func: a callable.

        The progress_func will be called multiple times with values between 0.0
        amd 1.0.
    """

    def __init__(
            self,
            source_path,
            target_path,
            band_count=1,
            progress_func=None,
    ):
        self.source_path = source_path
        self.target_path = target_path
        self.band_count = band_count
        self.progress_func = progress_func

        if path.exists(self.target_path):
            raise OSError("%s already exists." % self.target_path)

    def __enter__(self):
        """Open datasets.
        """
        self.source = gdal.Open(self.source_path, gdal.GA_ReadOnly)
        block_x_size, block_y_size = self.block_size
        options = ["compress=deflate", "blockysize=%s" % block_y_size]
        if block_x_size != self.raster_x_size:
            options += ["tiled=yes", "blockxsize=%s" % block_x_size]

        self.target = gdal.GetDriverByName("gtiff").Create(
            self.target_path,
            self.raster_x_size,
            self.raster_y_size,
            self.band_count,
            self.source.GetRasterBand(1).DataType,
            options=options,
        )
        self.target.SetProjection(self.projection)
        self.target.SetGeoTransform(self.geo_transform)
        self.target.GetRasterBand(1).SetNoDataValue(self.no_data_value)

        return self

    def __exit__(self, *args):
        """Close datasets.
        """
        self.source = None
        self.target = None

    @property
    def projection(self):
        return self.source.GetProjection()

    @property
    def geo_transform(self):
        return self.source.GetGeoTransform()

    @property
    def no_data_value(self):
        return self.source.GetRasterBand(1).GetNoDataValue()

    @property
    def raster_x_size(self):
        return self.source.RasterXSize

    @property
    def raster_y_size(self):
        return self.source.RasterYSize

    @property
    def block_size(self):
        return self.source.GetRasterBand(1).GetBlockSize()

    def __len__(self):
        block_size = self.block_size
        blocks_x = -(-self.raster_x_size // block_size[0])
        blocks_y = -(-self.raster_y_size // block_size[1])
        return blocks_x * blocks_y

    def partition(self):
        """Return generator of band_no, (xoff, xsize), (yoff, ysize) values.
        """
        def offset_size_range(stop, step):
            for start in range(0, stop, step):
                yield start, min(step, stop - start)

        # make y the outer loop, tiled tiff writing is much faster row-wise...
        raster_size = self.raster_y_size, self.raster_x_size
        block_size = self.block_size[::-1]
        generator = product(*map(offset_size_range, raster_size, block_size))

        total = len(self)
        for count, (y_part, x_part) in enumerate(generator, start=1):
            # ...and in the result put x before y
            yield x_part, y_part
            if self.progress_func is not None:
                self.progress_func(count / total)

    def convert_using(self, calculator, band):
        """Convert data writing it to tiff.

        Args:
            calculator (BaseCalculator): Calculator implementation instance
            band (int): Which band to write to.
        """
        no_data_value = self.no_data_value
        for (xoff, xsize), (yoff, ysize) in self.partition():
            # read
            values = self.source.ReadAsArray(
                xoff=xoff, yoff=yoff, xsize=xsize, ysize=ysize
            )
            indices = (yoff, xoff), (yoff + ysize, xoff + xsize)

            # calculate
            result = calculator(
                indices=indices,
                values=values,
                no_data_value=no_data_value,
            )

            # write - note GDAL counts bands starting at 1
            self.target.GetRasterBand(band + 1).WriteArray(
                array=result, xoff=xoff, yoff=yoff,
            )


class NetcdfConverter(GeoTIFFConverter):
    """Convert NetCDF4 according to the CF-1.6 standards."""

    def __init__(
            self,
            source_path,
            target_path,
            result_admin,
            calculation_steps,
            **kwargs
    ):
        kwargs["band_count"] = len(calculation_steps)
        super().__init__(source_path, target_path, **kwargs)

        self.ra = result_admin
        self.calculation_steps = calculation_steps

    def __enter__(self):
        """Open datasets"""
        self.source = gdal.Open(self.source_path, gdal.GA_ReadOnly)
        self.target = netCDF4.Dataset(self.target_path, "w")
        self._set_coords()
        self._set_time()
        self._set_meta_info()
        self._create_variable()

        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """Close datasets"""
        self.source = None
        self.target.close()

    def _set_meta_info(self):
        """Set meta info in the root group"""
        self.target.Conventions = "CF-1.6"
        self.target.institution = "3Di Waterbeheer"
        self.target.model_slug = self.ra.model_slug
        postfix = {"s1": "", "s1_max": " (using s1_max)"}[self.ra.variable]
        self.target.result_type = "Derived water depth" + postfix
        self.target.references = "http://3di.nu"

    def _set_time(self):
        """Set time"""

        self.target.createDimension("time", self.band_count)
        time = self.target.createVariable("time", "f4", ("time",))
        time.standard_name = "time"
        time.calendar = "standard"
        time.axis = "T"
        time.units = self.ra.get_time_units()
        time[:] = self.ra.get_timestamps(self.calculation_steps)

    def _set_coords(self):
        geotransform = self.source.GetGeoTransform()

        self.target.createDimension("y", self.raster_y_size)
        ycoords = self.target.createVariable("y", "f4", ("y",))

        # In CF-1.6 the coordinates are cell centers, while GDAL interprets
        # them as the upper-left corner.
        y_upper_left = geotransform[3] + geotransform[5] / 2
        ycoords[:] = np.arange(
            y_upper_left,
            y_upper_left + geotransform[5] * self.raster_y_size,
            geotransform[5]
        )
        ycoords.standard_name = "projection_y_coordinate"
        ycoords.long_name = "y coordinate of projection"
        ycoords.units = "m"
        ycoords.axis = "Y"

        self.target.createDimension("x", self.raster_x_size)
        xcoords = self.target.createVariable("x", "f4", ("x",))

        # CF 1.6 coordinates are cell center, while GDAL interprets
        # them as the upper-left corner.
        x_upper_left = geotransform[0] + geotransform[1] / 2
        xcoords[:] = np.arange(
            x_upper_left,
            x_upper_left + geotransform[1] * self.raster_x_size,
            geotransform[1]
        )
        xcoords.standard_name = "projection_x_coordinate"
        xcoords.long_name = "x coordinate of projection"
        xcoords.units = "m"
        xcoords.axis = "X"

        projection = self.target.createVariable(
            "projected_coordinate_system", "i4"
        )
        projection.EPSG_code = f"EPSG:{self.ra.epsg_code}"
        projection.epsg = self.ra.epsg_code
        projection.long_name = "Spatial Reference"
        projection.spatial_ref = osr.GetUserInputAsWKT(
            f"EPSG:{self.ra.epsg_code}"
        )  # for GDAL

    def _create_variable(self):
        water_depth = self.target.createVariable(
            "water_depth",
            "f4",
            ("time", "y", "x",),
            fill_value=-9999,
            zlib=True
        )
        water_depth.long_name = "water depth"
        water_depth.units = "m"
        water_depth.grid_mapping = "projected_coordinate_system"

    def convert_using(self, calculator, band):
        """Convert data writing it to netcdf4."""

        no_data_value = self.no_data_value
        for (xoff, xsize), (yoff, ysize) in self.partition():
            # read
            values = self.source.ReadAsArray(
                xoff=xoff, yoff=yoff, xsize=xsize, ysize=ysize
            )

            # calculate
            indices = (yoff, xoff), (yoff + ysize, xoff + xsize)
            result = calculator(
                indices=indices,
                values=values,
                no_data_value=no_data_value,
            )

            # write
            water_depth = self.target['water_depth']
            water_depth[band, yoff:yoff + ysize, xoff:xoff + xsize] = result


class ProgressClass:
    """ Progress function and calculation step iterator in one.

    Args:
        calculation_steps (list(int)): Calculation steps
        progress_func: a callable.

    The main purpose is iterating over the calculation steps, but inside the
    iteration progress_class() can be passed values between 0 and 1 to record
    the partial progress for a calculation step. The supplied `progress_func`
    will be called with increasing values up to and including 1.0 for the
    complete iteration.
    """
    def __init__(self, calculation_steps, progress_func):
        self.progress_func = progress_func
        self.calculation_steps = calculation_steps

    def __iter__(self):
        """ Generator of (band_no, calculation_step) """
        for band_no, calculation_step in enumerate(self.calculation_steps):
            self.current = band_no
            yield band_no, calculation_step
        del self.current

    def __call__(self, progress):
        """ Progress method for the current calculation step. """
        self.progress_func(
            (self.current + progress) / len(self.calculation_steps)
        )


class ResultAdmin:
    """
    args:
        gridadmin_path (str): Path to gridadmin.h5 file.
        results_3di_path (str): Path to (aggregate_)results_3di.nc file.

    Wraps either GridH5ResultAdmin or GridH5AggregateResultAdmin, based on the
    result type of the file at results_3di_path . Also has custom properties
    `result_type`, `variable` and `calculation_steps`.
    """

    def __init__(self, gridadmin_path, results_3di_path):
        with h5py.File(results_3di_path) as h5:
            self.result_type = h5.attrs['result_type'].decode('ascii')

        result_admin_args = gridadmin_path, results_3di_path
        if self.result_type == "raw":
            self._result_admin = GridH5ResultAdmin(*result_admin_args)
            self.variable = "s1"
            self.calculation_steps = self.nodes.timestamps.size
        else:
            self._result_admin = GridH5AggregateResultAdmin(*result_admin_args)
            self.variable = "s1_max"
            self.calculation_steps = self.nodes.timestamps[self.variable].size

    def get_timestamps(self, calculation_steps):
        if self.result_type == "raw":
            return self.nodes.timestamps[calculation_steps]
        else:
            return self.nodes.timestamps[self.variable][calculation_steps]

    def get_time_units(self):
        if self.result_type == "raw":
            return self.time_units.decode("utf-8")
        else:
            nc = self._result_admin.netcdf_file
            return nc['time_s1_max'].attrs['units'].decode('utf-8')

    def __getattr__(self, name):
        return getattr(self._result_admin, name)


calculator_classes = {
    MODE_COPY: CopyCalculator,
    MODE_NODGRID: NodGridCalculator,
    MODE_CONSTANT_S1: ConstantLevelCalculator,
    MODE_LINEAR_S1: LinearLevelCalculator,
    MODE_LIZARD_S1: LizardLevelCalculator,
    MODE_CONSTANT: ConstantLevelDepthCalculator,
    MODE_LINEAR: LinearLevelDepthCalculator,
    MODE_LIZARD: LizardLevelDepthCalculator,
}


def calculate_waterdepth(
    gridadmin_path,
    results_3di_path,
    dem_path,
    waterdepth_path,
    calculation_steps=None,
    mode=MODE_LIZARD,
    progress_func=None,
    netcdf=False,
):
    """Calculate waterdepth and save it as GeoTIFF.

    Args:
        gridadmin_path (str): Path to gridadmin.h5 file.
        results_3di_path (str): Path to (aggregate_)results_3di.nc file.
        dem_path (str): Path to dem.tif file.
        waterdepth_path (str): Path to waterdepth.tif file.
        calculation_steps (list(int)): Calculation step (default: [-1] (last))
        mode (str): Interpolation mode.
    """
    try:
        CalculatorClass = calculator_classes[mode]
    except KeyError:
        raise ValueError("Unknown mode: '%s'" % mode)

    result_admin = ResultAdmin(
        gridadmin_path=gridadmin_path, results_3di_path=results_3di_path,
    )

    # handle calculation step
    max_calculation_step = result_admin.calculation_steps - 1
    if calculation_steps is None:
        calculation_steps = [max_calculation_step]
    else:
        assert min(calculation_steps) >= 0
        assert max(calculation_steps) <= max_calculation_step, (
            "Maximum calculation step is '%s'." % max_calculation_step
        )

    # TODO remove at some point, newly produced gridadmins don't need it
    fix_gridadmin(gridadmin_path)

    progress_class = ProgressClass(
        calculation_steps=calculation_steps, progress_func=progress_func,
    )
    converter_kwargs = {
        "source_path": dem_path,
        "target_path": waterdepth_path,
        "progress_func": None if progress_func is None else progress_class,
    }
    if netcdf:
        converter_class = NetcdfConverter
        converter_kwargs['result_admin'] = result_admin
        converter_kwargs['calculation_steps'] = calculation_steps
    else:
        converter_class = GeoTIFFConverter
        converter_kwargs['band_count'] = len(calculation_steps)

    with converter_class(**converter_kwargs) as converter:
        calculator_kwargs_except_step = {
            "result_admin": result_admin,
            "dem_geo_transform": converter.geo_transform,
            "dem_shape": (converter.raster_y_size, converter.raster_x_size),
        }
        for band, calculation_step in progress_class:
            calculator_kwargs = {
                "calculation_step": calculation_step,
                **calculator_kwargs_except_step,
            }

            with CalculatorClass(**calculator_kwargs) as calculator:
                converter.convert_using(calculator=calculator, band=band)



# -*- coding: utf-8 -*-

from functools import reduce
from itertools import zip_longest
from math import ceil
from math import floor
from math import log

from scipy import ndimage
import numpy as np


def morton_array(shape):
    """
    Return array with Morton numbers.

    Inspired by:
    https://graphics.stanford.edu/%7Eseander/bithacks.html#InterleaveBMN
    """
    # determine the number of dimensions
    ndims = len(shape)

    # 1d compatibility
    if ndims == 1:
        return np.arange(shape[0])

    def bitcount(number):
        """ Return amount of bits used for in number """
        return int(ceil(log(number + 1, 2)))

    # feasbility check
    for i, j in enumerate(shape):
        # bit number assessment
        count = bitcount(j)                 # in the number
        count += (ndims - 1) * (count - 1)  # after spacing
        count += (ndims - 1) - i            # after shifting
        # numpy does not go higher than 64 bits currently
        if count > 64:
            raise ValueError('Too many bits needed for the computation')

    # generate list of zeros and masks
    ones = 1
    masks = []
    shifts = []
    pos = range(63, -1, -1)
    bmax = max(map(bitcount, shape))
    while ones < bmax:
        zeros = (ndims - 1) * ones
        shifts.append(zeros)
        period = ones + zeros
        masks.append(
            int(''.join('1' if i % period < ones else '0' for i in pos), 2),
        )
        ones *= 2

    # make indices and space them
    indices = [np.uint64(k) for k in np.ogrid[tuple(map(slice, shape))]]
    for i, (j, k) in enumerate(zip(shape, indices)):
        if j < 2:
            continue
        if j > 2:
            start = int(floor(log(bitcount(j) - 1, 2)))
        else:
            start = 0
        for n in range(start, -1, -1):
            k[:] = (k | k << shifts[n]) & masks[n]
        k <<= (ndims - 1) - i
    return reduce(np.bitwise_or, indices)


def get_morton_lut(array, no_data_value):
    """
    Return lookup table to rearrange an array of ints in morton order.

    :param array: 2D int array with a range of integers from 0 to no_data_value
    :param no_data_value: no data value that is excluded from rearrangement.

    The no_data_value does not have to be present in the array, but if it is,
    it does not get reordered by the lookup table (lut):
    lut[no_data_value] == no_data_value
    """
    # morton variables have underscores
    _array = morton_array(array.shape)
    _no_data_value = _array.max().item() + 1

    # make lookup from node to morton number
    index = np.arange(no_data_value + 1)
    lut1 = ndimage.minimum(_array, labels=array, index=index)
    lut1[no_data_value] = _no_data_value

    # make lookup from morton number back to node numbers
    lut2 = np.empty(_no_data_value + 1, dtype='i8')
    lut2[np.sort(lut1)] = index
    lut2[_no_data_value] = no_data_value

    # return the combined lookup table
    return lut2[lut1]


def group(array):
    """
    Return generator of arrays of indices to equal values.
    """
    order = array.argsort()
    _, index = np.unique(array[order], return_index=True)
    for start, stop in zip_longest(index, index[1:]):
        yield order[start:stop]


def analyze(x, y):
    """ Return (x_step, y_step) tuple.

    Return the smallest separation between points in the x-direction for points
    with the same y-coordinates and vice versa. That reveals the highest
    refinement level of the quadtree structure.
    """
    assert x.dtype == float
    assert y.dtype == float

    init = {'initial': np.inf}

    xs = min(np.diff(np.sort(x[i])).min(**init) for i in group(y))
    ys = min(np.diff(np.sort(y[i])).min(**init) for i in group(x))

    return None if np.isinf(xs) else xs, None if np.isinf(ys) else ys


def rasterize(points):
    """ Return (array, no_data_value) tuple.

    Rasterize the indices of the points in an array at the highest quadtree
    resolution. Note that points of larger squares in the quadtree also just
    occupy one cell in the resulting array, the rest of the cells get the
    no_data_value.
    """
    points = np.asarray(points, dtype=float)
    x, y = points.transpose()
    xs, ys = analyze(x, y)
    x1, y2 = x.min(), y.max()

    # get indices to land each point index in its own array cell
    j = np.int64(np.zeros_like(x) if xs is None else (x - x1) / xs)
    i = np.int64(np.zeros_like(y) if ys is None else (y2 - y) / ys)

    index = i, j
    no_data_value = len(points)
    ids = np.arange(no_data_value)

    values = np.full((i.max() + 1, j.max() + 1), no_data_value)
    values[index] = ids

    return values, no_data_value


def reorder(points, s1):
    """
    Return (points, s1) reordered to morton order.
    """
    array, no_data_value = rasterize(points)

    # array[lut] would have the ids in array in morton order
    lut = get_morton_lut(array=array, no_data_value=no_data_value)

    # the points need to be reordered such that rasterize(points[inv]) becomes
    # equal to lut[rasterize(points)] - in other words, for 'index value' a in
    # the raster to become 'index value' b in the raster, the index b in the
    # reordered points array must be occupied by the point from index a in the
    # old points array
    inv = np.arange(no_data_value)
    inv[lut[inv]] = inv.copy()  # may get bogus results without the copy

    return points[inv], s1[inv]



# -*- encoding: utf-8 -*-
from glob import glob
from os.path import basename
from os.path import splitext
from setuptools import find_packages
from setuptools import setup

version = '0.5'

with open('README.rst') as readme_file:
    readme = readme_file.read()

with open('CHANGES.rst') as changes_file:
    changes = changes_file.read()

install_requires = [
    "h5py",
    "h5netcdf",
    "numpy",
    "scipy",
    "threedigrid",
    # "gdal",  # assumed available via virtualenv --system-site-packages
]

tests_require = ["flake8", "pytest", "pytest-cov"]


setup(
    name='threedidepth',
    description="Calculate waterdepths for 3Di results.",
    version=version,
    author="Arjan Verkerk",
    author_email='arjan.verkerk@nelen-schuurmans.nl',
    license="BSD license",
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Natural Language :: English',
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        'Topic :: Scientific/Engineering',
    ],
    long_description=readme + '\n\n' + changes,
    python_requires='>=3.5',
    keywords=['threedidepth'],
    packages=find_packages('src'),
    package_dir={'': 'src'},
    py_modules=[splitext(basename(path))[0] for path in glob('src/*.py')],
    test_suite='test',
    url='https://github.com/nens/threedidepth',
    include_package_data=True,
    zip_safe=False,
    install_requires=install_requires,
    tests_require=tests_require,
    extras_require={"test": tests_require},
    entry_points={
        'console_scripts': [
            'threedidepth=threedidepth.commands:threedidepth',
        ],
    },
)



# -*- coding: utf-8 -*-

from unittest import mock
import sys

from threedidepth import commands


def test_command(tmpdir):
    depth_path = tmpdir.join("waterdepth.tif")
    depth_path.ensure(file=True)  # "touch" the file
    with mock.patch("threedidepth.commands.calculate_waterdepth") as wd:
        args = ["threedidepth,", "a", "b", "c", "d"]
        with mock.patch.object(sys, "argv", args):
            commands.threedidepth()
        wd.assert_called_with(
            gridadmin_path="a",
            results_3di_path="b",
            dem_path="c",
            waterdepth_path="d",
            calculation_steps=None,
            mode=commands.MODE_LIZARD,
            progress_func=None,
            netcdf=False
        )
        args.append("--constant")
        with mock.patch.object(sys, "argv", args):
            commands.threedidepth()
        wd.assert_called_with(
            gridadmin_path="a",
            results_3di_path="b",
            dem_path="c",
            waterdepth_path="d",
            calculation_steps=None,
            mode=commands.MODE_CONSTANT,
            progress_func=None,
            netcdf=False
        )


def test_command_with_multiple_steps(tmpdir):
    depth_path = tmpdir.join("waterdepth.tif")
    depth_path.ensure(file=True)  # "touch" the file
    with mock.patch("threedidepth.commands.calculate_waterdepth") as wd:
        args = ["threedidepth,", "a", "b", "c", "d", "--steps", "1", "2", "3"]
        with mock.patch.object(sys, "argv", args):
            commands.threedidepth()
        wd.assert_called_with(
            gridadmin_path="a",
            results_3di_path="b",
            dem_path="c",
            waterdepth_path="d",
            calculation_steps=[1, 2, 3],
            mode=commands.MODE_LIZARD,
            progress_func=None,
            netcdf=False
        )
